/*--------------------------------------------*/
/*        Written by Siddharth Agarwal        */
/*--------------------------------------------*/
/*--------------------------------------------*/
/*  An Nbody type program                     */
/*--------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define N 6

 
/*  The function that is used to compute the correlation  */
int correl(int a, int b)
{
  return (a*a + b*b);
}


main(int argc, char* argv[])

{ 
   int x[N], y[N];  /*  arrays containing the two sets of measurements  */       
   int n;    /* this tells us the no of things assigned to each processor  */
   int to;    /*   rank of process that the current process is sending to  */
   int from;    /*  rank of process from which current process receives data  */     
   int my_rank;  /*  rank of the current process  */     
   int local[N];    /*  correlation array for each processor initialized to zero  */
   int temp[N];    /*  temporary array used to receive y's from other processes  */
   int sum=0;  /*  used to sum up the elements of the correlation array for each processor  */
   int correlation=0;    /*the sum of all the correlation functions calculated by each process  */
   int p;  /*  the total number of processes  */
   int i,j,k;    /*  used for looping  */

   MPI_Comm grid_comm;     
   MPI_Status status;
   MPI_Init(&argc, &argv);
   MPI_Comm_size(MPI_COMM_WORLD, &p);
   MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

   n = N/p;       /*  each processor has N/p values  */      

   /* initializing the arrays*/
   for (i=0; i<n;i++) 
   {
    local[i] = 0;
    temp[i]=0;
   }
  
   /* assigning random variables to x & y*/
   for (i=0; i<n;i++) 
   {
     y[i] = my_rank*n + i; 
     x[i] = my_rank*n + i;
   }
       
     for (k=0; k<p; k++) 
     {	
       for ( j=0; j<n; j++)
	for ( i=0; i<n; i++)             
	  local[i] += correl(x[i], y[j]);
          
          to = (my_rank+1);
          from = my_rank-1;
	  if (to >= p)
          to = 0;
	  if (from < 0)
	  from = p-1;
          MPI_Sendrecv(&y, n, MPI_INT,to ,0, &temp, n, MPI_INT,from ,0, MPI_COMM_WORLD, &status);  
	  for(i = 0; i < n; i++)	
	   y[i] = temp[i];              
   }
          for(i = 0; i < n; i++)
	   sum += local[i];
			

/*a global reduction operation carried out to sum up the correlation functions obtained from all the processors  */
MPI_Reduce(&sum, &correlation, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

/*  Process 0 displays the correlation function obtained  */
if (my_rank == 0)
printf("\nThe correlation function is %d\n\n", correlation); 

MPI_Finalize();/* Closes the MPI session*/
}

